parser grammar Frank;

options
{
	language = 'CSharp';
	output = AST;
	tokenVocab = Inform6Lexer;
}
	
/* Inform's expression parser is context-sensitive: certain operators are
 * disallowed in certain contexts to work around ambiguities. Here we use
 * a dynamic scope, together with semantic predicates in the expression
 * rules, to disable the rules for those operators when necessary. */

scope Expr
{
	bool noComma;
	bool noFuncCall;
	bool noArrow;
	bool noSuper;
	bool noMinus;
}

program	:	expression;

/* Here we divide the keywords into sets. There's an added complication: some
 * keywords appear in more than one set. Here we simply create additional
 * sets to represent the intersections, like "directiveOrTraceKeywords", which
 * holds the keywords that appear in both the "directives" and "trace" sets. */

directiveKeywords
	:	ABBREVIATE | ARRAY | ATTRIBUTE | CONSTANT | END | ENDIF | EXTEND |
		FAKE_ACTION | GLOBAL | IFDEF | IFNDEF | IFNOT | IFV3 | IFV5 |
		IFTRUE | IFFALSE | IMPORT | INCLUDE | LINK | LOWSTRING | MESSAGE |
		NEARBY | RELEASE | SERIAL | SWITCHES | STATUSLINE | STUB |
		SYSTEM_FILE | TRACE | VERB | VERSION | ZCHARACTER;

traceKeywords
	:	ASSEMBLY | EXPRESSIONS | LINES | LINKER | OBJECTS | SYMBOLS |
		TOKENS | VERBS;

directiveOrTraceKeywords
	:	DICTIONARY;

traceOrMiscKeywords
	:	ON | OFF;

directiveSubKeywords
	:	ADDITIVE | ALIAS | BUFFER | CREATURE | ERROR | FATALERROR | FIRST |
		HELD | LAST | META | MULTI | MULTIEXCEPT | MULTIHELD |
		MULTIINSIDE | NOUN | ONLY | SCOPE | SCORE | SPECIAL |
		TABLE | TERMINATING | TIME | TOPIC | WARNING;

directiveSubOrSegmentKeywords
	:	PRIVATE | WITH;

directiveSubOrConditionOrSegmentKeywords
	:	HAS;

directiveOrSubKeywords
	:	REPLACE;

directiveOrSubOrSegmentKeywords
	:	CLASS;

miscKeywords
	:	A | ADDRESS | BOLD | CHAR | FIXED | FROM | NAME | NEAR | ROMAN |
		THE | TO | UNDERLINE;

directiveSubOrMiscKeywords
	:	NUMBER | REVERSE;

directiveOrMiscKeywords
	:	OBJECT | PROPERTY;

statementKeywords
	:	BOX | BREAK | CONTINUE | DO | ELSE | FONT | FOR | GIVE | IF |
		INVERSION | JUMP | MOVE | NEW_LINE | OBJECTLOOP | PRINT |
		PRINT_RET | QUIT | READ | REMOVE | RESTORE | RETURN | RFALSE |
		RTRUE | SAVE | SPACES | STYLE | SWITCH | UNTIL | WHILE;

directiveOrStatementKeywords
	:	DEFAULT;

directiveOrStatementOrMiscKeywords
	:	STRING;

conditionKeywords
	:	HASNT | IN | NOTIN | OFCLASS | OR | PROVIDES;

allKeywords
	:	directiveKeywords | traceKeywords | directiveSubKeywords |
		miscKeywords | statementKeywords | conditionKeywords |
		directiveOrTraceKeywords | traceOrMiscKeywords |
		directiveOrSubKeywords | directiveOrMiscKeywords |
		directiveOrStatementKeywords | directiveSubOrSegmentKeywords |
		directiveSubOrMiscKeywords | directiveOrSubOrSegmentKeywords |
		directiveSubOrConditionOrSegmentKeywords |
		directiveOrStatementOrMiscKeywords;

/* These rules allow us to match identifiers based on context, by excluding
 * the keyword sets that are active in a particular context. For example,
 * "directiveID" matches any word that isn't a directive keyword. */

directiveID
	:	ID | OPCODE | traceKeywords | directiveSubKeywords |
		miscKeywords | statementKeywords | conditionKeywords |
		traceOrMiscKeywords | directiveSubOrSegmentKeywords |
		directiveSubOrMiscKeywords |
		directiveSubOrConditionOrSegmentKeywords;

directiveSubID
	:	ID | OPCODE | directiveKeywords | traceKeywords |
		miscKeywords | statementKeywords | conditionKeywords |
		directiveOrTraceKeywords | traceOrMiscKeywords |
		directiveOrMiscKeywords | directiveOrStatementKeywords |
		directiveOrStatementOrMiscKeywords;

miscID	:	ID | OPCODE | directiveKeywords | traceKeywords |
		directiveSubKeywords | statementKeywords | conditionKeywords |
		directiveOrTraceKeywords | directiveOrSubKeywords | 
		directiveOrStatementKeywords | directiveSubOrSegmentKeywords |
		directiveOrSubOrSegmentKeywords |
		directiveSubOrConditionOrSegmentKeywords |
		directiveOrStatementOrMiscKeywords;

statementID
	:	ID | OPCODE | directiveKeywords | traceKeywords |
		directiveSubKeywords | miscKeywords | conditionKeywords |
		directiveOrTraceKeywords | traceOrMiscKeywords |
		directiveOrSubKeywords | directiveOrMiscKeywords |
		directiveSubOrSegmentKeywords | directiveSubOrMiscKeywords |
		directiveOrSubOrSegmentKeywords |
		directiveSubOrConditionOrSegmentKeywords;

conditionID
	:	ID | OPCODE | directiveKeywords | traceKeywords |
		directiveSubKeywords | miscKeywords | statementKeywords |
		directiveOrTraceKeywords | traceOrMiscKeywords |
		directiveOrSubKeywords | directiveOrMiscKeywords |
		directiveOrStatementKeywords | directiveSubOrSegmentKeywords |
		directiveSubOrMiscKeywords | directiveOrSubOrSegmentKeywords |
		directiveOrStatementOrMiscKeywords;

/* These rules exclude multiple keyword sets: "statementConditionID" matches
 * any word except a statement keyword or a condition keyword. */

statementConditionID
	:	ID | OPCODE | directiveKeywords | traceKeywords |
		directiveSubKeywords | miscKeywords |
		directiveOrTraceKeywords | traceOrMiscKeywords |
		directiveOrSubKeywords | directiveOrMiscKeywords |
		directiveSubOrSegmentKeywords |
		directiveSubOrMiscKeywords | directiveOrSubOrSegmentKeywords;

directiveSegmentID
	:	ID | OPCODE | traceKeywords | directiveSubKeywords |
		miscKeywords | statementKeywords | conditionKeywords |
		traceOrMiscKeywords | directiveSubOrMiscKeywords;

anyID	:	ID | OPCODE | allKeywords;

/* Sometimes, single- and double-quoted strings are interchangeable. */

anystr	:	STR | CHR | DICTWORD;

/* This rule could be used to check that a class name is actually in use. */

className
	:	validObjectName;

/* These rules validate identifiers by delegating to the relevant context-
 * sensitive rules for the contexts where the identifiers are used. */

validActionName
	:	anyID;

validAttrName
	:	directiveSubID;

validCustomGrammarToken
	:	directiveSubID;

validLabelName
	:	statementConditionID;

validObjectName
	:	directiveSegmentID;

validPropName
	:	directiveSubID;

validRoutineName
	:	directiveSubID;

validVarName
	:	statementConditionID;

num	:	DECNUM | HEXNUM | BINNUM;

/* Directives. */


/* Statements. */

/* Expressions. First, some rules to set the expression parsing context. */

actionExpr
scope Expr;
@init { $Expr::noFuncCall = true; }
	:	expression;

arrayExpr
scope Expr;
@init { $Expr::noMinus = true; $Expr::noFuncCall = true; }
	:	expression;

assemblyExpr
scope Expr;
@init { $Expr::noArrow = true; }
	:	expression;

conditionExpr
	:	normalExpr;

constantExpr
scope Expr;
@init { $Expr::noComma = true; }
	:	expression;

forInitExpr
scope Expr;
@init { $Expr::noSuper = true; }
	:	expression;

normalExpr
scope Expr;
	:	expression;

voidExpr:	normalExpr;

/* Now, rules for the various operators, from lowest to highest precedence.
 * Operators are usually left-associative. */

expression
	:	commaExpression;

/* , */

commaExpression
	:	assignmentExpression
			({!$Expr::noComma}?	(COMMA assignmentExpression)* )?;

/* =	(right-associative) */

assignmentExpression
	:	logicalExpression (EQUAL assignmentExpression)? ;

/* ~~ && || */

logicalExpression
	:	TILDETILDE* relationalExpression
			( (AMPAMP | PIPEPIPE) TILDETILDE* relationalExpression )* ;

/* == ~= > >= < <= has hasnt in notin provides ofclass	(non-associative) */

/* Note: "or" can only appear on the right side of a condition, so we skip
 * over it for the left side. */

relationalExpression
	:	additionExpression
			( ( EQUALEQUAL | TILDEEQUAL | RANGLE | RANGLEEQUAL | LANGLE | LANGLEEQUAL |
				HAS | HASNT | IN | NOTIN | PROVIDES | OFCLASS ) orExpression )?
	;

/* or */

orExpression
	:	additionExpression (OR additionExpression)* ;

/* + - */

additionExpression
	:	multiBitsExpression ( (PLUS | MINUS) multiBitsExpression )* ;

/* * / % & | ~ */

multiBitsExpression
	:	TILDE* arrayExpression
			( (STAR | SLASH | PERCENT | AMP | PIPE) TILDE* arrayExpression )*
	;

/* -> --> */

arrayExpression
	:	uminusExpression
			( {!$Expr::noArrow}?	ARROW uminusExpression
			| DARROW uminusExpression
			)*
	;

/* unary- */

uminusExpression
	:	( {!$Expr::noMinus}?	MINUS* )? incrementExpression;

/* ++ -- */

incrementExpression
	:	(PLUSPLUS | MINUSMINUS) propArrayExpression
	|	propArrayExpression ( (PLUSPLUS | MINUSMINUS)=>	(PLUSPLUS | MINUSMINUS) )?
	;

/* .& .# ..& ..# */

propArrayExpression
	:	functionExpression
			( (DOTAMP | DOTHASH | DOTDOTAMP | DOTDOTHASH) functionExpression )*
	;

/* function calls */

functionExpression
	:	propertyExpression
			( {!$Expr::noFuncCall}?
				((LPAREN)=>	LPAREN normalExpr RPAREN)* )?;

/* . .. */

propertyExpression
	:	superclassExpression ( (DOT | DOTDOT) superclassExpression)* ;

/* :: */

superclassExpression
	:	atom ( {!$Expr::noSuper}?	(COLONCOLON atom)+ )? ;

/* The smallest unit of an expression. Not *exactly* atomic: it can contain
 * a full expression inside parentheses, starting over at lowest precedence.
 * (Contextual restrictions are lifted for the inner expression.) */

atom	:	LPAREN normalExpr RPAREN
	|	num
	|	(HASHRDOLLAR | HASH)? conditionID
	;

/* Statements. */

//statement
//	:	label statement
//	|	HASH directive SEMI

/*	|	ATSIGN OPCODE assemblyExpr*
			(ARROW validVarName)?
			(QUESTION TILDE? (validLabelName | RFALSE | RTRUE))? SEMI*/

//	|	ATSIGN assemblyExpr* SEMI

//	|	voidExpr SEMI
//	;

//label	:	DOT validLabelName SEMI;
