lexer grammar Inform6Lexer;

options
{
	language = 'CSharp';
}

/* The OPCODE token is used to match Z-machine or Glulx opcode names. The
 * set of valid opcodes depends on which target Inform is compiling for. */

tokens
{
	OPCODE;
}

/* The lexer. */

fragment
Letter	:	'a'..'z' | 'A'..'Z' | '_';
fragment
Digit	:	'0'..'9';
fragment
HexDigit:	'0'..'9' | 'a'..'f' | 'A'..'F';
fragment
Space
	:	' ' | '\r' | '\t' | '\n';

WS	:	Space+				{ $channel = HIDDEN; };
LINE_COMMENT
	:	'!' ~('\n'|'\r')* '\r'? '\n'	{ $channel = HIDDEN; };

/* Numbers can be specified in decimal, hexadecimal, or binary. Here we
 * recognize them as separate token types; parser rules combine them later. */

DECNUM	:	Digit+;
HEXNUM	:	'$' HexDigit+;
BINNUM	:	'$$' ('0'..'1')+;

/* Double-quoted strings and dictionary words. Note the token names, STR and
 * CHR, which are not to be confused with the keywords STRING and CHAR. */

STR	:	'"' ~('"')* '"';
CHR	:	'\'' ~('\'') '\'';
DICTWORD:	'\'' ~('\'') ~('\'')+ '\''
	|	'#n$' (~Space)+
	;

ACTION	:	('##' | '#a$') Letter (Digit | Letter)*;

/* Operators and other punctuation. */


HASH	:	'#';
SEMI	:	';';
LBRACKET:	'[';
RBRACKET:	']';
EQUAL	:	'=';
COMMA	:	',';
PLUS	:	'+';
ARROW	:	'->';
DARROW	:	'-->';
LPAREN	:	'(';
RPAREN	:	')';
TILDE	:	'~';
AMP	:	'&';
PIPE	:	'|';
TILDETILDE
	:	'~~';
AMPAMP	:	'&&';
PIPEPIPE:	'||';
STAR	:	'*';
SLASH	:	'/';
PERCENT	:	'%';
RBRACE	:	'}';
ATSIGN	:	'@';
QUESTION:	'?';
LANGLE	:	'<';
RANGLE	:	'>';
EQUALEQUAL
	:	'==';
TILDEEQUAL
	:	'~=';
LANGLEEQUAL
	:	'<=';
RANGLEEQUAL
	:	'>=';
COLONCOLON
	:	'::';
COLON	:	':';
DOT	:	'.';
LBRACE	:	'{';
MINUS	:	'-';
PLUSPLUS:	'++';
MINUSMINUS
	:	'--';
DOTAMP	:	'.&';
DOTHASH	:	'.#';
DOTDOTAMP
	:	'..&';
DOTDOTHASH
	:	'..#';
DOTDOT	:	'..';
HASHRDOLLAR
	:	'#r$';

/* Keywords. Inform has many, many of these, and divides them into sets
 * that can be individually turned on and off during parsing. Inform only
 * recognizes a string as a keyword when the relevant set is enabled; in
 * other contexts, it can be used as an identifier. Here we recognize all
 * the keywords, all the time, and use parser rules to allow them as
 * identifiers when appropriate. */

A	:	'a' 'n'?;
ABBREVIATE
	:	'abbreviate';
ADDITIVE:	'additive';
ADDRESS	:	'address';
ALIAS	:	'alias';
ARRAY	:	'array';
ASSEMBLY:	'assembly';
ATTRIBUTE
	:	'attribute';
BOLD	:	'bold';
BOX	:	'box';
BREAK	:	'break';
BUFFER	:	'buffer';
CHAR	:	'char';
CLASS	:	'class';
CONSTANT:	'constant';
CONTINUE:	'continue';
CREATURE:	'creature';
DEFAULT	:	'default';
DICTIONARY
	:	'dictionary';
DO	:	'do';
ELSE	:	'else';
END	:	'end';
ENDIF	:	'endif';
ERROR	:	'error';
EXPRESSIONS
	:	'expressions';
EXTEND	:	'extend';
FAKE_ACTION
	:	'fake_action';
FATALERROR
	:	'fatalerror';
FIRST	:	'first';
FIXED	:	'fixed';
FONT	:	'font';
FOR	:	'for';
FROM	:	'from';
GIVE	:	'give';
GLOBAL	:	'global';
HAS	:	'has';
HASNT	:	'hasnt';
HELD	:	'held';
IFDEF	:	'ifdef';
IFFALSE	:	'iffalse';
IFNDEF	:	'ifndef';
IFNOT	:	'ifnot';
IFTRUE	:	'iftrue';
IFV3	:	'ifv3';
IFV5	:	'ifv5';
IMPORT	:	'import';
IF	:	'if';
IN	:	'in';
INCLUDE	:	'include';
INVERSION
	:	'inversion';
JUMP	:	'jump';
LAST	:	'last';
LINES	:	'lines';
LINK	:	'link';
LINKER	:	'linker';
LOWSTRING
	:	'lowstring';
MESSAGE	:	'message';
META	:	'meta';
MOVE	:	'move';
MULTI	:	'multi';
MULTIEXCEPT
	:	'multiexcept';
MULTIHELD
	:	'multiheld';
MULTIINSIDE
	:	'multiinside';
NAME	:	'name';
NEAR	:	'near';
NEARBY	:	'nearby';
NEW_LINE:	'new_line';
NOTIN	:	'notin';
NOUN	:	'noun';
NUMBER	:	'number';
OBJECT	:	'object';
OBJECTLOOP
	:	'objectloop';
OBJECTS	:	'objects';
OFCLASS	:	'ofclass';
OFF	:	'off';
ON	:	'on';
ONLY	:	'only';
OR	:	'or';
PRINT	:	'print';
PRINT_RET
	:	'print_ret';
PRIVATE	:	'private';
PROPERTY:	'property';
PROVIDES:	'provides';
QUIT	:	'quit';
READ	:	'read';
RELEASE	:	'release';
REMOVE	:	'remove';
REPLACE	:	'replace';
RESTORE	:	'restore';
RETURN	:	'return';
REVERSE	:	'reverse';
RFALSE	:	'rfalse';
ROMAN	:	'roman';
RTRUE	:	'rtrue';
SAVE	:	'save';
SCOPE	:	'scope';
SCORE	:	'score';
SERIAL	:	'serial';
SPACES	:	'spaces';
SPECIAL	:	'special';
STATUSLINE
	:	'statusline';
STRING	:	'string';
STUB	:	'stub';
STYLE	:	'style';
SWITCH	:	'switch';
SWITCHES:	'switches';
SYMBOLS	:	'symbols';
SYSTEM_FILE
	:	'system_file';
TABLE	:	'table';
THE	:	'the';
TIME	:	'time';
TERMINATING
	:	'terminating';
TO	:	'to';
TOKENS	:	'tokens';
TOPIC	:	'topic';
TRACE	:	'trace';
UNDERLINE
	:	'underline';
UNTIL	:	'until';
VERB	:	'verb';
VERBS	:	'verbs';
VERSION	:	'version';
WARNING	:	'warning';
WITH	:	'with';
WHILE	:	'while';
ZCHARACTER
	:	'zcharacter';

ID	:	Letter (Digit | Letter)*	{ if (opcodes.ContainsKey($text)) $type = OPCODE; }
	;
